"""
Author: "Rangana Warshamanage, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import argparse
import sys

cmdl_parser = argparse.ArgumentParser('EMDA commandline caller')
subparsers = cmdl_parser.add_subparsers(dest='command')

anyfsc = subparsers.add_parser('anyfsc')
anyfsc.add_argument('--map1',required=True, help='input map1 map')
anyfsc.add_argument('--map2',required=True, help='input map2 map')

halffsc = subparsers.add_parser('halffsc')
halffsc.add_argument('--h1',required=True, help='input map1 map')
halffsc.add_argument('--h2',required=True, help='input map2 map')

ccmask = subparsers.add_parser('ccmask')
ccmask.add_argument('--h1',required=True, help='input halfmap1 map')
ccmask.add_argument('--h2',required=True, help='input halfmap2 map')

lowpass = subparsers.add_parser('lowpass')
lowpass.add_argument('--map',required=True, help='input map (mrc/map)')
lowpass.add_argument('--res',required=True, type=float, help='lowpass resolution (A)')

power = subparsers.add_parser('power')
power.add_argument('--map',required=True, help='input map (mrc/map)')

applybfac = subparsers.add_parser('bfac')
applybfac.add_argument('--map',required=True, help='input map (mrc/map)')
applybfac.add_argument('--bfc',required=True, nargs='+', type=float, help='bfactor(s) to apply')
#applybfac.add_argument('--out',required=False, default=False, type=bool, help='if use write out maps')
applybfac.add_argument('--out', action='store_true', help='write out map')

map_resol = subparsers.add_parser('resol')
map_resol.add_argument('--h1',required=True, help='input halfmap1 map')
map_resol.add_argument('--h2',required=True, help='input halfmap2 map')

conv_mrc2mtz = subparsers.add_parser('mrc2mtz')
conv_mrc2mtz.add_argument('--map',required=True, help='input map (mrc/map)')
conv_mrc2mtz.add_argument('--out',required=True, help='output map (mtz)')

transform_map = subparsers.add_parser('transform')
transform_map.add_argument('--map',required=True, help='input map (mrc/map)')
transform_map.add_argument('--tra',required=False, default=[0.0, 0.0, 0.0], nargs='+', type=float, help='translation vec.')
transform_map.add_argument('--rot',required=False, default=0.0, type=float, help='rotation in deg')
transform_map.add_argument('--axr',required=False, default=[1,0,0], nargs='+', type=int, help='rotation axis')
transform_map.add_argument('--out',required=False, default='transformed.mrc', help='output map (mrc/map)')

conv_mtz2mrc = subparsers.add_parser('mtz2mrc')
conv_mtz2mrc.add_argument('--mtz',required=True, help='input map (mtz)')
conv_mtz2mrc.add_argument('--map',required=True, help='input map (mrc/map)')
conv_mtz2mrc.add_argument('--out',required=True, help='output map (mrc/map)')

resample_d = subparsers.add_parser('resample')
resample_d.add_argument('--map',required=True, help='input map (mrc/map)')
resample_d.add_argument('--pix',required=True, type=float, help='target pixel size (A)')
resample_d.add_argument('--out',required=False, default='resampled.mrc', help='output map name')

realspc = subparsers.add_parser('rcc')
realspc.add_argument('--h1',required=True, help='input halfmap1 map')
realspc.add_argument('--h2',required=True, help='input halfmap2 map')
realspc.add_argument('--mdl', required=False, help='Input model (map/mrc/mtz)')
realspc.add_argument('--knl', required=False, type=int, default=5, help='Kernel size (pixels)')

fourierspc = subparsers.add_parser('fcc')
fourierspc.add_argument('--h1',required=True, help='input halfmap1 map')
fourierspc.add_argument('--h2',required=True, help='input halfmap2 map')
fourierspc.add_argument('--knl', required=False, type=int, default=5, help='Kernel size (pixels)')

mapmodelfsc = subparsers.add_parser('mapmodelfsc')
mapmodelfsc.add_argument('--h1', required=True, help='input halfmap1 map')
mapmodelfsc.add_argument('--h2', required=True, help='input halfmap2 map')
mapmodelfsc.add_argument('--mdf', required=True, help='input full atomic model')
mapmodelfsc.add_argument('--md1', required=True, help='input halfmap1 atomic model')
mapmodelfsc.add_argument('--msk', required=False, help='input mask (mrc/map)')
mapmodelfsc.add_argument('--res', required=False, type=float, help='Resolution (A)')
mapmodelfsc.add_argument('--dim', required=False, nargs='+', type=np.int, help='map dim ')

mapoverlay = subparsers.add_parser('overlay')
mapoverlay.add_argument('--map',required=True, nargs='+', type=str, help='maplist for overlay')
mapoverlay.add_argument('--msk',required=False, default=None, nargs='+', type=str, help='masklist for overlay')
mapoverlay.add_argument('--tra',required=False, default=[0.0, 0.0, 0.0], nargs='+', type=float, help='translation vec.')
mapoverlay.add_argument('--rot',required=False, default=0.0, type=float, help='rotation in deg')
mapoverlay.add_argument('--axr',required=False, default=[1,0,0], nargs='+', type=int, help='rotation axis')
mapoverlay.add_argument('--ncy',required=False, default=10, type=int, help='number of fitting cycles')
mapoverlay.add_argument('--res',required=False, default=6, type=float, help='starting fit resol. (A)')

mapaverage = subparsers.add_parser('average')
mapaverage.add_argument('--map',required=True, nargs='+', type=str, help='maplist to average')
mapaverage.add_argument('--msk',required=False, default=None, nargs='+', type=str, help='masklist for maps')
mapaverage.add_argument('--tra',required=False, default=[0.0, 0.0, 0.0], nargs='+', type=float, help='translation vec.')
mapaverage.add_argument('--rot',required=False, default=0.0, type=float, help='rotation in deg')
mapaverage.add_argument('--axr',required=False, default=[1,0,0], nargs='+', type=int, help='rotation axis')
mapaverage.add_argument('--ncy',required=False, default=10, type=int, help='number of fitting cycles')
mapaverage.add_argument('--res',required=False, default=6, type=float, help='starting fit resol. (A)')

def anymap_fsc(args):
    from emda.emda_methods import twomap_fsc
    from emda import plotter
    res_arr, bin_fsc = twomap_fsc(args.map1, args.map2)
    plotter.plot_nlines(res_arr,
                        [bin_fsc],
                        'twomap_fsc.eps',
                        curve_label=["twomap_fsc"])

def halfmap_fsc(args):
    from emda.emda_methods import halfmap_fsc
    from emda import plotter
    res_arr, bin_fsc = halfmap_fsc(args.h1, args.h2)
    plotter.plot_nlines(res_arr,
                        [bin_fsc],
                        'halfmap_fsc.eps',
                        curve_label=["halfmap_fsc"])

def cc_mask(args):
    from emda.emda_methods import mask_from_halfmaps
    _ = mask_from_halfmaps(args.h1,args.h2)

def lowpass_map(args):
    from emda.emda_methods import lowpass_map
    lowpass_map(args.map,args.res)

def power_map(args):
    from emda.emda_methods import get_map_power
    from emda import plotter
    res_arr, power_spectrum = get_map_power(args.map)
    plotter.plot_nlines_log(res_arr,
                            [power_spectrum],
                            curve_label=["Power"],
                            mapname='map_power.eps')

def mapresol(args):
    from emda.emda_methods import estimate_map_resol
    resol = estimate_map_resol(args.h1, args.h2)
    print('Map resolution (A):', resol)

def mrc2mtz(args):
    import numpy as np
    from emda.iotools import read_map,write_3d2mtz
    uc,ar1,origin = read_map(args.map)
    hf1 = np.fft.fftshift(np.fft.fftn(ar1))
    write_3d2mtz(uc,hf1,outfile=args.out+'.mtz')

def mtz2mrc(args):
    from emda.iotools import read_map,write_mrc
    from emda.maptools import mtz2map
    uc,ar,origin = read_map(args.map)
    dat = mtz2map(args.mtz,ar.shape)
    outfile=args.out+'.mrc'
    write_mrc(dat,outfile,uc,origin)

def resample_data(args):
    from emda.emda_methods import read_map, resample_data, write_mrc
    uc, arr, org = read_map(args.map)
    new_arr = resample_data(args.pix, arr.shape, uc, arr)
    write_mrc(new_arr,args.out,uc)

def realsp_corr(args):
    from emda.emda_methods import realsp_correlation
    realsp_correlation(args.h1, args.h2, args.knl, args.mdl)

def fouriersp_corr(args):
    from emda.emda_methods import fouriersp_correlation
    fouriersp_correlation(args.h1, args.h2, args.knl)

def map_model_fsc(args):
    from emda.emda_methods import map_model_validate
    map_model_validate(args.h1, args.h2, args.mdf, args.md1, args.msk, args.dim, args.res)

def map_overlay(args):
    from emda.emda_methods import overlay_maps
    overlay_maps(args.map, args.msk, args.tra, args.rot, args.axr, args.ncy, args.res)

def map_transform(args):
    from emda.emda_methods import map_transform
    map_transform(args.map, args.tra, args.rot, args.axr, args.out)

def map_average(args):
    from emda.emda_methods import average_maps
    average_maps(args.map, args.msk, args.tra, args.rot, args.axr, args.ncy, args.res)

def apply_bfac(args):
    from emda.emda_methods import apply_bfactor_to_map
    all_maps = apply_bfactor_to_map(args.map,args.bfc,args.out)




def main(command_line=None):
    args = cmdl_parser.parse_args(command_line)
    if args.command == 'anyfsc':
        anymap_fsc(args)
    if args.command == 'halffsc':
        halfmap_fsc(args)
    if args.command == 'ccmask':
        cc_mask(args)
    if args.command == 'lowpass':
        lowpass_map(args)
    if args.command == 'power':
        power_map(args)
    if args.command == 'resol':
        mapresol(args)
    if args.command == 'mrc2mtz':
        mrc2mtz(args)
    if args.command == 'mtz2mrc':
        mtz2mrc(args)
    if args.command == 'resample':
        resample_data(args)
    if args.command == 'rcc':
        realsp_corr(args)
    if args.command == 'fcc':
        fouriersp_corr(args)
    if args.command == 'mapmodelfsc':
        map_model_fsc(args)
    if args.command == 'overlay':
        map_overlay(args)
    if args.command == 'average':
        map_average(args)
    if args.command == 'transform':
        map_transform(args)
    if args.command == 'bfac':
        apply_bfac(args)


if __name__ == '__main__':
    main()


